# **FLAME CLMC Information Model Specification**

This document describe the configuration and monitoring specification for cross-layer management and control within the FLAME platform. All information measured by the CLMC aims to improve management and control decisions made by the platform against defined performance criteria such as increasing QoE or reducing costs. 

## **Overview**

This section provides an overview of the FLAME CLMC information model considering the following elements:

* Media Service Information
* Configuration
* Monitoring
* Information Security
* Privacy


### Media Service (https://gitlab.it-innovation.soton.ac.uk/mjb/flame-clmc/issues/2)

The FLAME architecture defines a media services as "An Internet accessible service supporting processing, storage and retrieval of content resources hosted and managed by the FLAME platform". A media service consists of 1 or more media components (also known as Service Functions) that together are composed to create an overall Service Function Chain. SFs are realised through the instantiation of virtual machines (or containers) based on resource management policy. Multiple VMs may be instantiated for each SF to create surrogate SFs to balance load and deliver against performance targets.

Media services are described using a template structured according to the TOSCA specification (http://docs.oasis-open.org/tosca/TOSCA/v1.0/TOSCA-v1.0.html). A TOSCA template includes all of the information needed for the FLAME orchestrator to instantiate a media service. This includes all SF's, links between SFs server resource configuration information. The Alpha version of the FLAME platform is based on the current published TOSCA specification. Future developments will extend the TOSCA specification (known as TOSCA++) to meet FLAME requirements such higher level KPIs and location-based constraints.

The current TOSCA template provides the initial structure of the Media Service information model through specified service and resource configuration. Within this structure, system components are instantiated whose runtime characteristics are measured to inform management processes. Measurements relate to individual SF's as well as aggregated measurements structured according the context of deployment (e.g. media service, platform, etc). Measurements are made by monitoring processes deployed with system components.

The media information model in relation to the high-level media service lifecycle is shown in the diagram below. The lifecycle includes processes for packaging, orchestration and management/control. Each stage in the process creates context for decisions and measurements within the next stage of the lifecycle. Packaging creates the context for orchestration, orchestration creates the context for service instantiation, etc. In the diagam, the green concepts provide the context for filtering and queries whilst the yellow concepts are the measurement data providing runtime characteristics

![FLAMEContext](/docs/image/flame-context.jpg)

The primary measurement point in the model is the VM/Container instance as this is the realisation of computational processes running on the platform. A VM/Container is one or more process running on a physical or virtual host with ports connecting to other VM/Container instances over network links. The VM/Container has measurement processes running to capture different views on the VM/Container including the network, host, and service. The acquisition of these different views of the VM/Container together are a key element of the cross-layer information required for management and control.  The measurements about a VM/Container are captured by different processes running on the VM or container but are brought together by comon context allowing the information to be integrated, correlated and analysed. 

### Configuration (https://gitlab.it-innovation.soton.ac.uk/mjb/flame-clmc/issues/3)

Configuration information describes the structure of the system over time. Each system component has a configuration lifecycle that defines configuration states and transistions between states.

Configuration information can include:

* Capacity (e.g. servers, networks)
* Topology (e.g. nodes and links)
* Resource Allocation (e.g. cpus, memory, data IO, etc)
* SF State (e.g. port up, port down, sf placed, sf booted)
* Media Service (software)

### Monitoring (https://gitlab.it-innovation.soton.ac.uk/mjb/flame-clmc/issues/8)

Monitoring measures the behaviour of the system overtime including metrics associated with usage and performance. Measurements are made within the context of a configuration. 

Usage monitoring information can include:

* network resource usage
* host resource usage
* service usage

Performance monitoring information can include:

* cpu/sec
* throughput
* response time
* etc.

### Information Security (https://gitlab.it-innovation.soton.ac.uk/mjb/flame-clmc/issues/25) (TBC Stephen Phillips)

*to be completed*

### Data Subject (https://gitlab.it-innovation.soton.ac.uk/mjb/flame-clmc/issues/24) (TBC Stephen Phillips)

*to be completed*

## **Measurement Model**

### General 

The measurement model is based on a time-series model defined by TICK stack from influxdata called the line protocol. The protocol defines a format for measurement samples which together can be combined to create series.

`<measurement>[,<tag-key>=<tag-value>...] <field-key>=<field-value>[,<field2-key>=<field2-value>...] [unix-nano-timestamp]`

Each series has:

* a name "measurement"
* 0 or more tags for configuration context
* 1 or more fields for the measurement values
* a timestamp.

The model is used to report both configuration and monitoring data. In general, tags are used to provide configuration context for measurement values stored in fields. The tags are structured to provide queries by KPIs and dimensions defined in the FLAME architecture. 

Tags are automatically indexed by InfluxDB. Global tags can be automatically inserted by contexualised agents collecting data from monitoring processes. The global tags used across different measurements are a key part of the database design. Although, InfluxDB is schemaless database allowing arbirtary measurement fields to be stored (e.g. allowing for a media component to have a set of specific metrics), using common global tags allows the aggregation of measurements across time with common context.

Although similar to SQL influx is not a relational database and the primary key for all measuremetns is time. Further schema design recommendations can be found here: https://docs.influxdata.com/influxdb/v1.4/concepts/schema_and_data_layout/

### Temporal Measurements (TBC Simon Crowle)

Monitoring data must have time-stamp values that are consistent and sycnrhonised across the platform. This means that all VMs hosting SFs should have a synchronised system clock, or at least (and more likely) a means by which an millisecond offset from the local time can be retrieved so that a 'platform-correct' time value can be calculated.

Describe approaches to integrate temporal measurements, time as a primary key, etc.

### Spatial Measurements (TBC Simon Crowle)

Location can be provided in two forms: labelled (tag) and numeric (longitude and latitude as digitial degrees). Note that the location label is likely to be a _global tag_. 

Tag location

| location | loc_long | loc_lat |
| --- | --- | --- |
| DATACENTRE_1 | 0 | 0 |

A SF media transcoder is placed in a lamp-post. It has no means to obtain GPS coordinates but has a _location_label_ provided to it as a VM environment variable. It provides zeros in the longitude and latitude. In subsequent data analysis we can search for this SF by location label.

GPS coordination location

| location_label | location_long | location_lat |
| --- | --- | --- |
| LAMP_1 | 50.842715 | -0.778276 |

A SF that is a proxy to a user attached to a NAP running in street lamp post LAMP_1. Here we have knowledge both of the logical location of the service and also the fine-grained, dynamic position of the service user.

Note that tags are always strings and cannot be floats, therefore log and lat will always be stored as a measurement field. 

Integrating and analysing location measurements

If tags are used then measurements of GPS coordinates will need to be translated into tag based approximation. For example, if a user device is tracking location information then for that to be combined with a server location the GPS coordinate needs to be translated.

Matching on tags is limited to matching and potentially spatial hierarchies (e.g. country.city.street). Using a coordiante system allows for mathatical functions to be developed (e.g. proximity functions)

## **Decision Context**

Monitoring data is collected to support service design, management and control decisions. The link between decisions and data is through queries and rules applied to contextual information stored with measurement values.  

![MeasurementContext](/docs/image/measurement-context.jpg)

Every measurement has a measurement context. The context allows for time-based series to be created according to a set of query criteria which are then be processed to calculate statistical data over the desired time-period for the series. For example, in the following query the measurement is avg_response_time, the context is “service A” and the series are all of the data points from now minus 10 minutes.

`find avg response time for service A between over the last 10 minutes`

To support this query the following measurement would be created:

`service_response,service_id=(string) response_time=(float) timestamp`

Designing the context for measurements is an important step in the schema design. This is especially important when measurements from multiple monitoring sources need to be integrated and processed to provided data for queries and decision. The key design principles adopted include:

* identify common context across different measurements
* where possible use the same identifiers and naming conventions for context across different measurements
* organise the context into hierarchies that are automatically added to measurements during the collection process

![CommonContext](/docs/image/common-measurement-context.jpg)

The following figure shows the general structure approach for two measurements A and B. Data points in each series have a set of tags that shares a common context and have a specific context related to the measurement values.

![FLAMEMeasurements](/docs/image/flame-measurements.jpg)

The measurement model considers three views on the VM/Container instance with field values covering:

* service: specific metrics associated within the SF (either media component or platform component) 
* network: data usage TX/RX, latency, jitter, etc.
* host: cpu, storage, memory, storage I/O, etc

All of the measurements on a specific VM/Container instance share a common context that includes tag values:

* sfc – an orchestration template
* sfc_instance – an instance of the orchestration template
* sf_package – a SF type
* sf_instance – an instance of the SF type
* vm_instance – an authoritive copy of the SF instance
* server – a physical or virtual server for hosting VM instances
* location – the location of the server
	
By including this context with service, network and host measurements it is possible to support a wide range of temporal queries associated with SFC’s. By adopting the same convention for identifiers it is possible to combine measurements across service, network and host to create new series that allows exploration of different aspects of the VM instance, including cross-layer queries.

Give a worked example across service and network measurements based on the mpeg-dash service

* Decide on the service management decisions and time scales
* Decide on the measurements of interest that are needed to make the decisions
* Decide how measurements are calculated from a series of one or more other measurements 
* Decide on time window for the series and sample rate
* Decide on interpolation approach for data points in the series

Discuss specific tags

## **Architecture**

### General ###
The monitoring model uses an agent based approach with hierarchical aggregation used as required for different time scales of decision making. The general architecture is shown in the diagram below.

![AgentArchitecture](/docs/image/agent-architecture.jpg)

To monitoring a SF an agent is deployed on each of the container/VM implementing a SF. The agent is deployed by the orchestrator when the SF is provisioned. The agent is configured with a set of input plugins that collect measurements from the three viewpoints of network, host and service. The agent is configured with a set of global tags that are inserted for all measurements made by the agent on the host.

Telegraf offers a wide range of integration with relevant monitoring processes.

* Telegraf Existing Plugins for common services, relevant plugins include
 * Network Response https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/net_response: could be used to performance basic network monitoring
 * nstat https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/nstat : could be used to monitor the network
 * webhooks https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/webhooks: could be used to monitor end devices
 * prostat https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/procstat: could be used to monitor containers
 * SNMP https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/snmp: could be used to monitor flows
 * systat https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/sysstat: could be used to monitor hosts

Telegraf offers a wide range of integration for 3rd party monitoring processes:

* Telegraf AMQP: https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/amqp_consumer
* Telegrapf http json: https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/httpjson
* Telegraf http listener: https://github.com/influxdata/telegraf/tree/release-1.5/plugins/inputs/http_listener 
* Telegraf Bespoke Plugin: https://www.influxdata.com/blog/how-to-write-telegraf-plugin-beginners/

The architecture considers hierarchical monitoring and scalability, for example, AMQP can be used to buffer monitoring information whilst InfluxDB can be used to provide intermediate aggregation points when used with Telegraf input and output plugin. 

### Integration with FLIPS Monitoring

FLIPS offers a scalable pub/sub system for distributing monitoring data. The architecture is described in the POINT monitoring specification https://drive.google.com/file/d/0B0ig-Rw0sniLMDN2bmhkaGIydzA/view. Some observations can be made

* MOOSE and CLMC provide similar functions in the architecture, the CLMC will not have access to MOOSE but will need to subscribe to data points provided by FLIPS
* The APIs for Moly and Blackadder are not provided therefore it's not possible to critically understand the correct implementation approach for agents and monitoring data distribution
* Individual datapoints  need to be aggregated into measurements 
* It's likely that we'll have to use the blackadder API for distribution of monitoring data, replacing messaging systems such as RabbitMQ with all buffering and pub/sub deployed on the nodes themselves rather than a central service. 

There are a few architectural choices. The first below uses moly as an integration point for monitoring processes via a Telegraf output plugin with data inserted into influx using a blackadder API input plugin on another Telegraf agent running on the CLMC. In this case managing the subscriptions to nodes and data points is difficult. In addition, some data points will be individual from FLIPS monitoring whilst others will be in line protocol format from Telegraf. For the FLIPS data points a new input plugin would be required to aggregate individual data points into time-series measurements. 

![FLIPSAgentArchitecture](/docs/image/flips-monitoring-architecture.jpg)

The second (currently preferred) choice only sends line protocol format over the wire. Here we develop telegraf input and output plugins for blackadder benefiting from the scalable nature of the pub/sub system rather than introducing RabbitMQ as a central server. In this case the agent on each node would be configured with input plugins for service, host and network . We'd deploy a new Telegraf input plugin for FLIPS data points on the node's agent by subscribing to blackadder locally and then publish the aggregated measurement using the line protocol back over blackadder to the CLMC. FLIPS can still publish data to MOOSE as required. 

![FLIPSAgentArchitecture](/docs/image/flips-monitoring-architecture2.jpg)

The pub/sub protocol still needs some work as we don't want the CLMC to have to subscribe to nodes as they start and stop. We want the nodes to register with a known CLMC and then start publishing data to the CLMC according to a monitoring configuration (e.g. sample rate, etc). So we want a "monitoring topic" that nodes publish to and that the CLMC can pull data from. This topic is on the CLMC itself and note the nodes. Reading the FLIPS specification it seems that this is not how the nodes current distribute data, although could be wrong

## Configuration Measurement Summary

|Decision Context|Measurement|Description
|---|---|---|
|Capacity|host_resource|the compute infrastructure allocation to the platform|
|Capacity|network_resource|the network infrastructure allocation to the platform|
|Platform|topology_manager|specific metrics tbd|
|Media Service|sfc_config|specific metrics tbd|
|Media Service|sf_config|specific metrics tbd|
|Media Service|vm_host_config|compute resources allocated to a VM|
|Media Service|net_port_config|networking constraints on port on a VM|

*Need to refer to TOSCA here*

## Monitoring (usage and performance) Measurement Summary

|Decision Context|Measurement|Description
|---|---|---|
|Platform|nap_data_io|nap data io at byte, ip and http levels|
|Platform|nap_fqdn_perf|fqdn request rate and latency|
|Platform|orchestrator|specific metrics tbd|
|Platform|clmc|specific metrics tbd|
|Media Service|cpu_usage|vm metrics|
|Media Service|disk_usage|vm metrics|
|Media Service|disk_IO|vm metrics|
|Media Service|kernel_stats|vm metrics|
|Media Service|memory_usage|vm metrics|
|Media Service|process_status|vm metrics|
|Media Service|swap_memory_usage|vm metrics|
|Media Service|system_load_uptime|vm metrics|
|Media Service|net_port_io|vm port network io and error at L2|
|Media Service|service|vm service perf metrics|


## Capacity 

Capacity measurements measure the size of the infrastructure slice available to the platform that can be allocated on demand to tenants.

Common tags

* slice_id – an idenfication id for the tenant infrastructure slice within openstack

**host_resource**

The *host_resource* measurement measures the wholesale host resources available to the platform that can be allocated to media services.

`host_resource,slice_id="",server_id="",location="" cpus=(integer),memory=(integer),storage=(integer) timestamp`

**network_resource**

network_resource measures the overall capacity of the network available to the platform for allocation to tenants. There are currently no metrics defined for this in the FLIPS monitoring specification, although we can envisage usage metrics such as bandwidth being part of this measurement.

`network_resource,slice_id="",network_id="", bandwidth=(integer),X=(integer),Y=(integer),Z=(integer) timestamp`

## Platform

Platform measurements measure the configuration, usage and performance of platform components.

**topology_manager**

**nap**

nap measurements are the platforms view on IP endpoints such as user equipment and services. A NAP is therefore the boundary of the platform. NAP also measures aspects of multicast performance 

NAP multicast metrics that require further understanding

Fields

* CHANNEL_AQUISITION_TIME_M 
* CMC_GROUP_SIZE_M

* What is the group id for CHANNEL_AQUISITION_TIME_M and how can this be related to FQDN of the content?
* what is the predefined time interval for CMC_GROUP_SIZE_M?
* How are multicast groups identified? i.e. "a request for FQDN within a time period", what's the content granularity here?

NAP data usage measurement

`nap_data_io,node_id="",ip_version="" <fields> timestamp`

Fields

* RX_BYTES_HTTP_M
* TX_BYTES_HTTP_M 
* RX_PACKETS_HTTP_M 
* TX_PACKETS_HTTP_M
* RX_BYTES_IP_M
* TX_BYTES_IP_M 
* RX_BYTES_IP_MULTICAST_M
* TX_BYTES_IP_MULTICAST_M 
* RX_PACKETS_IP_MULTICAST_M 
* TX_PACKETS_IP_MULTICAST_M 

NAP service request and response metrics

`nap_fqdn_perf,<common_tags>,cont_nav=FQDN <fields> timestamp`

Fields

* HTTP_REQUESTS_FQDN_M
* NETWORK_FQDN_LATENCY

**clmc**

## Media Service 

Media service measurements measure the configuration, usage and performance of media service instances deployed by the platform.

### Service Function Chain

**sfc_config**

tbd

**sf_config**

tbd

**sfc**

Aggregate measurement derived from VM/container measurements, most likely calculated using a continuous query of a specific time interval

**sf**

Aggregate measurement derived from VM/container measurements, most likely calculated using a continuous query of a specific time interval

**sfc_instance**

Aggregate measurement derived from VM/container measurements, most likely calculated using a continuous query of a specific time interval

**sf_instance**

Aggregate measurement derived from VM/container measurements, most likely calculated using a continuous query of a specific time interval

### VM/Container Measurements

VM/Container Measurements measure the configuration, usage and performance of VM/Container instances deployed by the platform within the context of a media service.

Common tags

* sfc – an orchestration template
* sfc_i – an instance of the orchestration template
* sf_pack – a SF package identifier indicating the type and version of SF
* sf_i – an instance of the SF type
* vm_i – an authoritive copy of the SF instance (node_id)
* server – a physical or virtual server for hosting VM instances
* location – the location of the server

#### Network Measurements

**net_port_config**

network config is concerned with any network io allocation/constraints for network rx/tx

`net_port_config,<common_tags>,port_state="",port_id="" <fields> timestamp`

Possible Fields

* RX_USAGE
* TX_USAGE
* RX_THROUGHPUT
* TX_THROUGHPUT

Specific tags
* port_state
* port_id

**net_port_io**

`network_io,<common_tags>,port_id="" <fields> timestamp`

Fields

* PACKET_DROP_RATE_M
* PACKET_ERROR_RATE_M
* RX_PACKETS_M 
* TX_PACKETS_PORT_M
* RX_BYTES_PORT_M 
* TX_BYTES_PORT_M 

Note that RX_PACKETS_M seems to have inconsistent naming convention. 

#### VM Host Measurements

SF Host Resource Measurements measures the host resources allocated to a service function deployed by the platform. All measurements have the following global tags to allow the data to be sliced and diced according to dimensions.

**vm_host_config**

The resources allocated to a VM/Container

`res_alloc,<global-tags> cpu,memory,storage timestamp`

Specific tags

* node_state

**cpu_usage**

[[inputs.cpu]]

**disk_usage**

[[inputs.disk]]

**disk_IO**

[[inputs.diskio]]

**kernel_stats**

[[inputs.kernel]]

**memory_usage**

[[inputs.mem]]

**process_status**

[[inputs.processes]]

**swap_memory_usage**

[[inputs.swap]]

**system_load_uptime**

[[inputs.system]]

#### Service Measurements

**<prefix>_service_config**

Fields
* <tbd>

Specific Tags
* service_state

**<prefix>_service_perf**

Each SF developed will measure service specific usage and performance measurements. The following are provided as examples or common service metrics.

`service,<global_tags>,cont_nav="",cont_rep="",user="" <fields> timestamp`

Fields (examples)

* request_rate
* response_time
* peak_response_time
* error_rate
* throughput

Specific Tags

* cont_nav: the content requested
* cont_rep: the content representation requested
* user: the pseudonym of an individual user or a user classification

# MISC Measurements and Questions

The following data points require further analysis

* CPU_UTILISATION_M: likely to be replaced by other metrics provided directly by Telegraf plugins
* END_TO_END_LATENCY_M: not clear what this measurement means, so needs clarification
* BUFFER_SIZES_M: needs clarification 
* RX_PACKETS_IP_M: is this just NAP or all Nodes
* TX_PACKETS_IP_M: is this just NAP or all Nodes

The following fields need further analysis as they seem to relate to core ICN, most likely fields/measurements related to platform components

* FILE_DESCRIPTORS_TYPE_M 
* MATCHES_NAMESPACE_M
* PATH_CALCULATIONS_NAMESPACE_M 
* PUBLISHERS_NAMESPACE_M 
* SUBSCRIBERS_NAMESPACE_M 

The following fields relate to CID which I don't understand but jitter is an important metric so we need to find out.

* PACKET_JITTER_CID_M
* RX_BYTES_CID_M 
* TX_BYTES_CID_M 

Some questions

* Can a single value of jitter (e.g. avg jitter) be calculated from the set of measurements in PACKET_JITTER_CID_M message? What is the time period for the list of jitter measurements?
* What does CID  mean? consecutive identical digits

#### Link Measurements

Links are established between VM/container instances, need to discuss what measurements make sense. Also the context for links could be between media services, therefore a link measurement should be within the platform context and NOT the media service context. Need a couple of scenarios to work this one out. 

**link_config**

Link Tags

* link_name
* link_id
* source_node_id
* destination_node_id
* link_type
* link_state

**link_perf**

link perf is measured at the nodes, related to end_to_end_latency. Needs further work.

# Other Issues

**Trust in measurements**

If the agent is deployed in a VM/container that a tenant has root access then a tenant could change the configuration to fake measurements associated with network and host in an attempt gain benefit. This is a security risk. Some ideas include

* Deploy additional agents on hosts rather than agents to measure network and VM performance. Could be hard to differentiate between the different SFs deployed on a host
* Generate a hash from the agent configuration file that's checked within the monitoring message. Probably too costly and not part of the telegraf protocol
* Use unix permissions (e.g. surrogates are deployed within root access to them)